#!/usr/bin/env ruby
require 'set'
require 'pathname'
SCORE_DIR = File.dirname(Pathname.new(File.expand_path(__FILE__)).realpath)
require SCORE_DIR + '/lib/metaify'
require SCORE_DIR + '/lib/shell_escape'

def get_num_refs(hyp, refs, expected_multiplier)
  hyp_lines = hyp.size
  refs_lines = refs.size
  throw "Number disagreement: #{hyp_lines} hypotheses and #{refs_lines} references" if refs_lines % hyp_lines != 0
  num_refs = refs_lines / hyp_lines
  throw "Expected #{expected_multiplier} reference per sentence, got #{num_refs}" if expected_multiplier and  num_refs != expected_multiplier
  num_refs
end

class Output
  def initialize(prefix, delete_temp, delete_perm)
    @prefix = prefix.clone
    if File.directory?(@prefix) then
      @prefix += '/' if  @prefix[-1] != '/'[0]
    else
      @prefix += '.' if File.exists?(@prefix)
    end
    @delete_temp = delete_temp
    @delete_perm = delete_perm
    @temps = []
    @perms = []
  end
  def name(suffix)
    @prefix + suffix
  end
  #Name a temporary file that will be deleted if @delete_temp
  def temp(suffix)
    ret = name(suffix)
    @temps << ret
    ret
  end
  #Name a permanent file
  def perm(suffix)
    ret = name(suffix)
    @perms << ret
    ret
  end
  def record_perm(full_name)
    @perms << full_name
  end

  def finish(print_results)
    $stdout.puts File.new(@prefix + "scores").read if print_results
    if @delete_temp then
      @temps.each do |t|
        File.unlink(t)
      end
    end
    if @delete_perm then
      @perms.each do |p|
        File.unlink(p)
      end
    end
  end
end

class Hypothesis
  attr_reader :file_name, :lines
  def initialize(file_name, output)
    @file_name = file_name
    @lines = File.new(@file_name).readlines
  end
end

class Reference
  attr_reader :laced_name, :laced_lines, :number
  def initialize(type, arg, hyp_count, output)
    #Goal: get @laced_name = name of a file with interlaced refs and @laced_lines the lines of this file.  
    #Optionally @number is the number of references so that an error message can be printed of the hypothesis is a fraction of this.  
    case type
    when :single then
      @number = 1
      @laced_name = arg
      @laced_lines = File.new(@laced_name).readlines
    when :laced then
      @laced_name = arg
      @laced_lines = File.new(@laced_name).readlines
    when :seq then
      lines = File.new(arg).readlines
  	  throw "There are #{hyp_count} hypotheses and #{lines.size} total references" unless lines.size % hyp_count == 0
  	  @number = lines.size / hyp_count
  	  @laced_lines = []
  	  hyp_count.times do |i|
        @number.times do |j|
  			  line = lines[i+j*hyp_count]
  			  @laced_lines << line
  		  end
  	  end
      build_laced_file(output, @laced_lines)
    when :multiple then
      @number = arg.size
      files = arg.map do |ref|
        File.new(ref)
      end
      @laced_lines = []
      loop do
        lines = files.map { |f| f.gets }
        compact = lines.compact
        break if compact.size == 0
        throw "Unequal number of lines" if compact.size != files.size
        lines.each do |l|
          @laced_lines << l
        end
      end
      build_laced_file(output, @laced_lines)
    else
      throw "Bad type #{type}"
    end
    throw "There are #{hyp_count} hypotheses and #{laced_lines.size} total references" unless laced_lines.size % hyp_count == 0
    number = laced_lines.size / hyp_count
    throw "Size of references is #{number} times number of hypotheses, but #{@number} specified" if @number and @number != number
    @number = number
  end
  def build_laced_file(output, content)
    @laced_lines = content.map { |l| l.strip }
    @laced_name = output.temp("ref.laced")
    file = File.new(@laced_name, "w")
    @laced_lines.each { |l| file.puts l }
    file.close
  end
  def range(segment)
    ((segment * number)..(segment * number + number - 1))
  end
end

class ScoreRequest
  attr_reader :output, :hyp, :ref, :cased, :no_norm, :language, :task, :individual
  def initialize(output, hyp, ref, cased, no_norm, language, task, individual)
    @output = output
    @hyp = hyp
    @ref = ref
    @cased = cased
    @no_norm = no_norm
    @language = language
    @task = task
    @individual = individual
  end
end

def format_scores(scores, formats)
  formats.zip(scores).map do |paired|
    sprintf(paired[0], paired[1])
  end.join(' ')
end

def write_close(name, content)
  f = File.new(name, "w")
  f.puts content
  f.close
  nil
end

#Takes a ScoreRequest
def score(request)
  metric = MetaMetric.new
  write_close(request.output.perm("scores"),
      format_scores(
        metric.summary_score(request),
        metric.summary_format(request)))    
  if request.individual then
    formats = metric.individual_format(request)
    scores = metric.individual_score(request)
    out_indiv = File.new(request.output.perm("individual"), "w")
    zipped = formats.zip(scores)
    request.hyp.lines.size.times do |i|
      out_indiv.puts(zipped.map do |z|
        sprintf(z[0], z[1][i])
      end.join(' '))
    end
    out_indiv.close
  end
end

def parse_args(names)
  var = nil
  parsed = {}
  process = true
  ARGV.each do |arg|
    if arg[0..1] == "--" and process then
      name = arg[2..-1]
      #-- ends argument processing
      if name == "" then
        process = false
        next
      end
      throw "Bad argument #{arg}" unless names.include?(name)
      throw "Duplicate argument #{arg}" if parsed[name]
      var = Array.new
      parsed[name] = var
      next
    end
    throw "First argument must begin with --" unless var
    var << arg
  end
  parsed
end


HEADER = "BLEU  NIST  TER   METEO Preci Recal Leng"
USAGE=__FILE__ + "
--hyp hyp.txt
[--ref ref.txt] [--refs ref1.txt ref2.txt] [--refs-laced ref-interlaced.txt]
[--output output_prefix] [--individual] [--print-header] [--print]
[--keep-temp] [--delete-results]
Scores using BLEU NIST TER METEOR METEOR_Precision METEOR_Recall Length in that order.
Aligned to output spacing, these scores are:
#{HEADER}

HYPOTHESIS
The hypothesis file is plaintext one sentence per line with no HTML character entities.
--hyp passes files as is.

REFERENCE
Specify the reference using exactly one of --ref, --refs, --refs-laced, or --refs-seq
All reference files are plaintext with no HTML character entity.
--ref means a single reference, one per line.  Same as --refs except this
  checks for one argument.  Also the same as --refs-laced except this checks
  for one reference.
--refs means multiple reference files are specified, where the nth line of each
  is a reference for line n of the hypothesis.
--refs-laced means that the references for the first segment appear at the top
  followed by references for the second segment, etc.  Further, the number of
  references for each sentence is constant.  This is the reference format
  accepted by Meteor's plaintext mode.  
--refs-seq means that the references are in sequence, i.e. first references for 
  all segments are followed by second references for all segments, etc. 

PARAMETERS
--cased activates case sensitivity for BLEU, NIST, and TER.  
  METEOR is only impacted if --no-norm is also specified.  
  The default is case insensitive.
--language sets the language, as accepted by METEOR.  BLEU and NIST get this
  set in the SGML field, which appears to have no impact on scoring.  TER
  doesn't care. 
--task is the task for METEOR: af, rank, or hter.  Default is hter.  
--no-norm disables normalization for BLEU and METEOR.  TER still does its thing.

OUTPUT
--output specifies a prefix for output.  This defaults to the hypothesis file
  with a period appended.  If a directory is specified, a / is appended if it
  does not appear already.
--individual creates a file with segment-level scores.  Currently these are
  TER, METEOR, and Length.  
--print-header prints the column headers.  This may be used alone.  
--print prints the scores to stdout.
--keep-temp will leave temporary files such as bogus NIST SGML files around.
--delete-results deletes result files, in which case you probably want --print too.
"

if ARGV.empty? or ARGV == ["--help"] or ARGV == ["-h"] then
  $stdout.puts USAGE
  exit 0
end

begin
  parsed = parse_args(Set.new(["hyp-tok", "hyp-detok", "hyp", "ref", "refs", "refs-laced", "refs-seq", "output", "cased", "keep-temp", "delete-results", "print-header", "print", "language", "task", "individual", "no-norm"]))
  ["cased", "no-norm", "print-header", "print", "keep-temp", "delete-results"].each do |name|
    throw "Expected no argument for #{name}" if parsed[name] and parsed[name].size != 0
  end
  #Special case: print-header may be called alone.  
  if parsed.key?("print-header") then
    $stdout.puts HEADER
    exit 0 if parsed.size == 1
  end
  throw "hyp-tok was a bad idea.  Run your own detokenizer before scoring." if parsed["hyp-tok"]
  ["hyp", "hyp-detok", "ref", "refs-laced", "refs-seq", "output", "language", "task"].each do |name|
    throw "Expected one argument for #{name}" if parsed[name] and parsed[name].size != 1
  end
  throw "Specify one of hyp or hyp-detok" unless [parsed["hyp"], parsed["hyp-detok"]].compact.size == 1
  throw "Specify one of ref, refs, refs-laced or refs-seq" unless [parsed["ref"], parsed["refs"], parsed["refs-laced"], parsed["refs-seq"]].compact.size == 1
rescue NameError => bang
  $stderr.puts USAGE
  $stderr.puts
  $stderr.puts "ERROR with your command:"
  $stderr.puts bang.name
  exit 2
end

hypothesis_file = (parsed["hyp"] || parsed["hyp-detok"])[0]
unless File.exists?(hypothesis_file) then
  $stderr.puts "No such file #{hypothesis_file}"
  exit 1
end
output = Output.new(parsed.key?("output") ? parsed["output"][0] : hypothesis_file, (not parsed.key?("keep-temp")), parsed.key?("delete-results"))

File.new(output.perm("args"), "w").puts(escape_shell_array([__FILE__] + ARGV))

hyp = Hypothesis.new(hypothesis_file, output)
ref = if parsed.key?("ref") then
  Reference.new(:single, parsed["ref"][0], hyp.lines.size, output)
elsif parsed.key?("refs-laced") then
  Reference.new(:laced, parsed["refs-laced"][0], hyp.lines.size, output)
elsif parsed.key?("refs-seq") then
  Reference.new(:seq, parsed["refs-seq"][0], hyp.lines.size, output)
else
  Reference.new(:multiple, parsed["refs"], hyp.lines.size, output)
end
language = parsed.key?("language") ? parsed["language"][0] : "en"
task = parsed.key?("task") ? parsed["task"][0] : "hter"
individual = parsed.key?("individual")
score(ScoreRequest.new(output, hyp, ref, parsed.key?("cased"), parsed.key?("no-norm"), language, task, individual))

output.finish(parsed.key?("print"))
